/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001-2005  by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@gmx.de
 *    Borg Enders  borg@borgsoft.de
 *
 ********************************************************************/

#ifdef USE_NETWORK

#include "constants.h"

#include "connection.h"
#include "server.h"
#include "types.h"
#include "exceptions.h"

#include "gnet/connection.h"

#include "../misc/setting.h"
#include "../ui/ui.h"

#include "../utils/string.h"

namespace Network {

  /**
   ** constructor
   **
   ** @param	server		the corresponding server
   ** @param	address		address to connect to
   ** @param	port		port to connect to
   **
   ** @return	-
   **
   ** @author	Diether Knof
   **
   ** @version	0.7.0
   **/
  Connection*
    Connection::new_(Server& server,
		     string const& address, unsigned const port)
    {
      try {
	return new GNet::Connection(server, address, port);
      } catch (ConnectionFailureException const&) {
	return NULL;
      }
      return NULL;
    } // static Connection* Connection::new_(Server& server, string address, unsigned port)

  /**
   ** constructor
   **
   ** @param	server	the corresponding server
   **
   ** @return	-
   **
   ** @author	Diether Knof
   **
   ** @version	0.7.0
   **
   ** @todo	all
   **/
  Connection::Connection(Server& server) :
    server_(&server),
    name_(),
    address_(),
    port_(Server::STANDARD_PORT),
#ifdef POSTPHONED
    accepted_(false)
#else
      accepted_(true)
#endif
  {
    return ;
  } // Connection::Connection(Server server)

  /**
   ** destructor
   **
   ** @param	-
   **
   ** @return	-
   **
   ** @author	Diether Knof
   **
   ** @version	0.7.0
   **/
  Connection::~Connection()
  {
  } // Connection::~Connection()

  /**
   ** -> result
   **
   ** @param	-
   **
   ** @return	the corresponding server
   **
   ** @author	Diether Knof
   **
   ** @version	0.7.0
   **/
  Server&
    Connection::server()
    {
      return *this->server_;
    } // Server& Connection::server()

  /**
   ** a text has been received
   **
   ** @param	text	received text
   **
   ** @return	-
   **
   ** @author	Diether Knof
   **
   ** @version	0.7.0
   **/
  void
    Connection::received(string const& text)
    {
      if (!this->accepted())
	return ;

      ::ui->network_traffic(*this, TRAFFIC_INCOMING, text);
      this->analyse_and_act(text);

      return ;
    } // void Connection::received(string const& text)

  /**
   ** close this connection
   **
   ** @param	-
   **
   ** @return	-
   **
   ** @author	Diether Knof
   **
   ** @version	0.7.0
   **
   ** @todo	all
   **/
  void
    Connection::close()
    {
      this->server().delete_connection(this);

      return ;
    } // void Connection::close()

  /**
   ** send the data of this freedoko instance:
   ** name
   **
   ** @param	-
   **
   ** @return	-
   **
   ** @author	Diether Knof
   **
   ** @version	0.7.0
   **/
  void
    Connection::send_self_data()
    {
      this->send("name: " + ::setting(Setting::NAME));

      return ;
    } // void Connection::send_self_data()

  /**
   ** analyes the text and acts accordinatly
   **
   ** @param	text	received text
   **
   ** @return	whether the text was accepted
   **
   ** @author	Diether Knof
   **
   ** @version	0.7.0
   **/
  bool
    Connection::analyse_and_act(string const& text)
    {
      if (!this->accepted())
	return false;

      /* 
       *
       * Network
       *   ping			reply with 'pong'
       *   pong			'ping' reply
       *   ping: %dnumber%	reply with 'pong: %dnumber'
       *   pong: %dnumber%	'ping' reply
       *
       * Other
       *   name: %sname%	new name of the connection
       *   chat: %stext%	chattet text
       *   request: %stype%	request data of the type (p.e. self data)
       *
       *
       */

      { // without parameter
	{ // Network
	  if (text == "ping") {
	    this->send("pong");
	    return true;
	  } if (text == "pong") {
	    // ToDo: check, whether this is a reply of a ping
	    return true;
	  } // if (text == ...)
	} // Network
      } // without parameter

      size_t const colon_pos = text.find(':');
      if (colon_pos != string::npos) {
	// one parameter
	string const word(text, 0, colon_pos);
	string const param(text,
			   ((text.size() > colon_pos)
			    && (text[colon_pos + 1] == ' '))
			   ? colon_pos + 2
			   : colon_pos + 1);

	// Network
	if (word == "ping") {
	  // reply with 'pong: %n+1%'
	  istringstream istr(param);
	  unsigned n;
	  istr >> n;
	  if (istr.fail())
	    return false;

	  this->send("pong: " + DK::Utils::String::to_string(n));

	  return true;

	} else if (word == "pong") {
	  // a reply of a ping
	  istringstream istr(param);
	  unsigned n;
	  istr >> n;
	  if (istr.fail())
	    return false;

	  // ToDo: check, whether this is a reply of a ping

	  return true;
	} else if (word == "name") {
	  // change the name of this connection
	  this->name_ = param;
	  ::ui->network_connection_changed(*this);
	  return true;

	} else if (word == "chat") {
	  // the player has chattet
	  ::ui->chat(this->name(), param, false);
	  return true;

	} else if (word == "request") {
	  // the player requests some data
	  if (param == "self data") {
	    this->send_self_data();
	    return true;
	  } else {
	    return false;
	  }
	} // if (word == ...)
      } // one parameter

      return false;
    } // bool Connection::analyse_and_act(string const& text)

} // namespace Network

#endif // #ifdef USE_NETWORK
